<?php

namespace Drupal\imago;

use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Session\AccountProxyInterface;
use Drupal\file\FileInterface;
use Drupal\media\Entity\Media;
use Drupal\token\Token;
use GuzzleHttp\Client;

/**
 * Fetch and save data from ContentPlatform.
 */
class ImagoManager {

  /**
   * The entity manager.
   *
   * @var \Drupal\Core\Entity\EntityManagerInterface
   */
  protected $entityTypeManager;

  protected $nodeStorage;
  protected $mediaStorage;

  protected $fieldConfigStorage;

  /**
   * Service to retrieve token information.
   *
   * @var \Drupal\token\Token
   */
  protected $token;

  /**
   * Current user.
   *
   * @var \Drupal\Core\Session\AccountProxyInterface
   */
  protected $currentUser;

  /**
   * Settings for stream wrappers.
   *
   * @var array
   */
  protected $settings = [];

  /**
   * Constructs a EntityCreateAccessCheck object.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_manager
   *   The entity manager.
   * @param \Drupal\Core\Session\AccountProxyInterface $current_user
   *   Current user account.
   * @param \Drupal\token\Token $token
   *   Service to retrieve token information.
   */
  public function __construct(EntityTypeManagerInterface $entity_manager, AccountProxyInterface $current_user, Token $token) {
    $this->entityTypeManager = $entity_manager;
    $this->nodeStorage = $this->entityTypeManager->getStorage('node');
    $this->mediaStorage = $this->entityTypeManager->getStorage('media');
    $this->fieldConfigStorage = $this->entityTypeManager->getStorage('field_config');
    $this->currentUser = $current_user;
    $this->token = $token;
  }

  /**
   * Fetch data from Image Search API.
   */
  public function connectSearch($filters = []) {
    $query_config = \Drupal::config('imago.settings');
    $config = $query_config->get('api');

    $content_uri = $config['query']['api_url'];
    $options = [
      'headers' => [
        'x-api-user' => $config['api_user'],
        'x-api-key' => $config['api_key'],
      ],
      'body' => json_encode($filters),
    ];
    $response = $this->validateResponse($content_uri, $options);
    $source_json = json_decode($response->getBody()->getContents(), TRUE);

    if (!empty($source_json)) {
      return $source_json[1]['pictures'];
    }
  }

  /**
   * Download image from Imago by imago id.
   *
   * @param int $imago_id
   *   Image-imago.de image id.
   *
   * @return bool|\Drupal\Core\Entity\EntityInterface
   *   Media entity.
   */
  public function download(int $imago_id) {
    $query_config = \Drupal::config('imago.settings');
    $api = \Drupal::state()->get('imago.api_query.' . $this->currentUser->id());
    $config = $query_config->get('api');
    $filters = [
      "pictureid" => $imago_id,
      "db" => $api[$imago_id]['database'],
      "res" => "9",
    ];
    $content_uri = $config['download']['api_url'];
    $options = [
      'headers' => [
        'x-api-user' => $config['api_user'],
        'x-api-key' => $config['api_key'],
      ],
      'body' => json_encode($filters),
    ];

    $response = $this->validateResponse($content_uri, $options);
    $file = $response->getBody()->detach();

    return $this->createImageEntity($imago_id, $file, $query_config->get('media'));
  }

  /**
   * Validate http response and return data.
   *
   * @param string $url
   *   Requested url.
   * @param array $options
   *   Request options.
   *
   * @return bool|array
   *   Array with JSON data, or FALSE.
   *
   * @throws \Exception
   */
  private function validateResponse(string $url, array $options = []) {
    try {
      /** @var \GuzzleHttp\Client $client */
      $client = new Client();
      return $client->post($url, $options);
    }
    catch (\Exception $e) {
      watchdog_exception('imago', $e, $e->getMessage());
      return FALSE;
    }
  }

  /**
   * Verify if image exist in our system, or create new one.
   *
   * @param int $imago_id
   *   Imago id.
   * @param string $bundle
   *   Media bundle.
   *
   * @return bool|\Drupal\Core\Entity\EntityInterface
   *   Media entity.
   */
  public function getEntity(int $imago_id, string $bundle = 'media') {
    $properties = [];
    if ($bundle === 'file') {
      $properties['filename'] = 'imago_' . $imago_id;
    }
    elseif ($bundle === 'media') {
      $properties['name'] = 'imago_' . $imago_id;
      $properties['domain_access'] = $this->getDomainAccess();
    }
    $existing_entities = $this->entityTypeManager->getStorage($bundle)->loadByProperties($properties);
    if (!empty($existing_entities)) {
      return reset($existing_entities);
    }

    return NULL;
  }

  /**
   * Create image entity.
   *
   * @param int $imago_id
   *   Source image data.
   * @param resource $stream
   *   Source image stream.
   * @param string $bundle
   *   Media bundle for creating image.
   *
   * @return bool|\Drupal\Core\Entity\EntityInterface
   *   Media entity if image was created.
   *   False if image can't be created.
   *
   * @throws \Exception
   */
  private function createImageEntity(int $imago_id, $stream, string $bundle) {
    $file_id = $this->saveToStorage($imago_id, $bundle, 'field_media_image', $stream);
    $data = \Drupal::state()->get('imago.api_query.' . $this->currentUser->id());
    if ($file_id) {
      $image_media = Media::create([
        'name' => 'imago_' . $imago_id,
        'bundle' => $bundle,
        'uid' => $this->currentUser->id(),
        'field_media_image' => [
          'target_id' => $file_id,
          'alt' => substr($data[$imago_id]['caption'], 0, 512),
          'width' => $data[$imago_id]['width'],
          'height' => $data[$imago_id]['height'],
        ],
        'field_description' => [
          'value' => $data[$imago_id]['caption'],
          'format' => 'unprocessed',
        ],
        'field_media_source' => $data[$imago_id]['source'],
        'domain_access' => $this->getDomainAccess(),
        'field_image_height' => $data[$imago_id]['height'],
        'field_image_width' => $data[$imago_id]['width'],
        'picture_id' => $imago_id,
        'created' => $data[$imago_id]['created'],
      ]);

      if ($image_media->save() === SAVED_NEW) {
        return $image_media;
      }
      return FALSE;
    }

    throw new \Exception(t('Image can\'t be saved.'));
  }

  /**
   * Save image to storage.
   *
   * @param int $imago_id
   *   Image-imago.de image id.
   * @param string $bundle
   *   Media bundle.
   * @param string $image_field
   *   Image field name.
   * @param string $stream
   *   Data stream.
   *
   * @return int|false
   *   File ID or FALSE in case of error.
   *
   * @throws \Exception
   */
  private function saveToStorage(int $imago_id, $bundle, $image_field, $stream) {
    $existing_file = $this->getEntity($imago_id, 'file');
    if ($existing_file instanceof FileInterface) {
      return $existing_file->id();
    }

    $field_instance = $this->fieldConfigStorage->load('media.' . $bundle . '.' . $image_field);
    $file_directory = $this->token->replace($field_instance->getSetting('file_directory'));
    $uri_scheme = $field_instance->getSetting('uri_scheme');
    $destination = $uri_scheme . '://' . $file_directory . '/' . $imago_id . '.jpg';
    $file = file_save_data($stream, $destination);
    if ($file instanceof FileInterface) {
      $file->setFilename('imago_' . $imago_id);
      $save = $file->save();
      if ($file && ($save === SAVED_NEW || $save === SAVED_UPDATED)) {
        return $file->id();
      }
      else {
        throw new \Exception(t('File can\'t be saved.'));
      }
    }
    return FALSE;
  }

  /**
   * Get parent node domain access.
   *
   * @return string
   *   Domain machinename.
   */
  private function getDomainAccess() {
    $parent_query = \Drupal::request()->query->all();
    $nid = explode('/', $parent_query['original_path'])[2];
    $domain_access = 'default';
    if (is_numeric($nid)) {
      $node = $this->nodeStorage->load($nid);
      $domain_access = $node->get('field_domain_access')->getString();
    }
    return $domain_access;
  }

}
